#include "x86.h"
#include "elf.h"

/*****************************************************************************
 * memos-boot
 *
 *   Set up the memos kernel, which is loaded at address 0x100000 (1 MB).
 *
 *   The kernel's entry point is stored at the top of its stack, at 0x1FFFFC.
 *
 *   Don't worry about understanding this bootloader!
 *
 *****************************************************************************/

/**********************************************************************
 * This a dirt simple boot loader, whose sole job is to boot
 * an ELF kernel image from the first IDE hard disk.
 *
 * DISK LAYOUT
 *  * This program (bootstart.S and passwdos-boot.c) is the bootloader.
 *    It should be stored in the first sector of the disk.
 * 
 *  * The 2nd through 33rd sectors hold the image for the first
 *    process, the password printer (passwdos-p.c).
 *
 *  * These process images must be in ELF executable format.
 *
 * BOOT UP STEPS	
 *  * When the CPU boots it loads the BIOS into memory and executes it.
 *
 *  * The BIOS intializes devices, sets of the interrupt routines, and
 *    reads the first sector of the boot device (e.g., hard-drive) 
 *    into memory and jumps to it.
 *
 *  * Assuming this boot loader is stored in the first sector of the
 *    hard-drive, this code takes over.
 *
 *  * Control starts in bootstart.S, which sets up protected mode,
 *    and a stack so C code then run, then calls bootmain().
 *
 *  * bootmain() in this file takes over, reads in the two process images,
 *    and jumps to the first image (the printer).
 *
 **********************************************************************/

#define SECTORSIZE	512
#define PAGESIZE	4096
#define ELFHDR		((struct Elf *) 0x10000) // scratch space

void readsect(void *addr, uint32_t sectornum);
void readseg(uint32_t va, uint32_t filesz, uint32_t memsz, uint32_t offset);

void
bootmain(void)
{
	struct Proghdr *ph, *eph;
	uint32_t *stackptr;

	// read 1st page off disk
	readseg((uint32_t) ELFHDR, PAGESIZE, PAGESIZE, 0);

	// is this a valid ELF?
	if (ELFHDR->e_magic != ELF_MAGIC)
		return;

	// load each program segment (ignores ph flags)
	ph = (struct Proghdr*) ((uint8_t *) ELFHDR + ELFHDR->e_phoff);
	eph = ph + ELFHDR->e_phnum;
	for (; ph < eph; ph++)
		readseg(ph->p_va, ph->p_filesz, ph->p_memsz, ph->p_offset);

	// store the entry point from the ELF header into the stack
	stackptr = (uint32_t *) (0x00080000 - 4);
	*stackptr = ELFHDR->e_entry & 0xFFFFFF;

	// jump to the kernel
	__asm __volatile("movl $0x0007FFFC, %esp; ret");
}

// Read 'filesz' bytes at 'offset' from kernel into virtual address 'va'.
// Might copy more than asked
void
readseg(uint32_t va, uint32_t filesz, uint32_t memsz, uint32_t offset)
{
	uint32_t end_va;

	va &= 0xFFFFFF;
	end_va = va + filesz;
	memsz += va;
	
	// round down to sector boundary
	va &= ~(SECTORSIZE - 1);

	// translate from bytes to sectors, and kernel starts at sector 1
	offset = (offset / SECTORSIZE) + 1;

	// read sectors
	while (va < end_va) {
		readsect((uint8_t*) va, offset);
		va += SECTORSIZE;
		offset++;
	}

	// clear bss segment
	while (end_va < memsz)
		*((uint8_t*) end_va++) = 0;
}

void
waitdisk(void)
{
	// wait for disk reaady
	while ((inb(0x1F7) & 0xC0) != 0x40)
		/* do nothing */;
}

void
readsect(void *dst, uint32_t offset)
{
	// wait for disk to be ready
	waitdisk();

	outb(0x1F2, 1);		// count = 1
	outb(0x1F3, offset);
	outb(0x1F4, offset >> 8);
	outb(0x1F5, offset >> 16);
	outb(0x1F6, (offset >> 24) | 0xE0);
	outb(0x1F7, 0x20);	// cmd 0x20 - read sectors

	// wait for disk to be ready
	waitdisk();

	// read a sector
	insl(0x1F0, dst, SECTORSIZE/4);
}

