#include "memos-kern.h"
#include "x86struct.h"
#include "x86.h"
#include "x86mem.h"

/*****************************************************************************
 * memos-kern
 *
 *   This is the memos's kernel.
 *   It sets up process control blocks for the 4 applications, then runs
 *   them in some schedule.
 *   It also contains 
 *
 *****************************************************************************/

// MEMOS INITIAL PHYSICAL MEMORY LAYOUT
//
//  +-------------- Base Memory --------------+
//  v                                         v
// +-----+--------------------+----------------+--------------------+---------/
// |     | Kernel      Kernel |       :    I/O | App 0        App 0 | App 1
// |     | Code + Data  Stack |  ...  : Memory | Code + Data  Stack | Code ...
// +-----+--------------------+----------------+--------------------+---------/
// 0  0x40000              0x80000 0xA0000 0x100000             0x140000
//                                         == PROCESS_START

process_t processes[NPROCESSES];	// Process descriptor array
process_t *current_proc;		// The currently running process

void run(process_t *process) __attribute__((noreturn));


/*****************************************************************************
 * start
 *
 *   Initialize the hardware and process control blocks, then run
 *   the first process.
 *
 *****************************************************************************/

void
start(void)
{
	int i;
	pagedirectory_t initial_pgdir = 0;

	// Set up hardware (memos-x86.c)
	segments_init();
	interrupt_controller_init();

	// Set up memory
	memory_labels_init();

	// Show physical memory on console
	console_clear();
	show_physical_memory();

	// Set up process descriptors (the processes[] array)
	for (i = 0; i < NPROCESSES; i++) {
		process_t *proc = &processes[i];

		// Load the process
		process_init(proc);
		proc->p_pid = i;
		proc->p_pgdir = initial_pgdir;
		process_load(proc);

		// Set up stack pointer:
		// %esp register should point to the top of the stack page
		proc->p_registers.reg_esp = PROCESS_START + (i+1) * PROCESS_SIZE;
		// Allocate stack page
		page_alloc(proc, proc->p_registers.reg_esp - PAGESIZE);

		// Mark process as runnable
		proc->p_runnable = 1;
	}

	// Switch to the first process.  The context_switch_to_pid() function
	// context switches to a particular process descriptor, which causes
	// that process to start running.
	run(&processes[0]);

	// Should never get here!
	while (1)
		/* do nothing */;
}



/*****************************************************************************
 * trap
 *
 *   This is the weensy interrupt and system call handler.
 *   The current handler handles 4 different system calls (two of which
 *   do nothing), plus the clock interrupt.
 *
 *   Note that we will never receive clock interrupts while in the kernel.
 *
 *****************************************************************************/

uint32_t seed = 982192143;

void
trap(registers_t *reg)
{
	// Save the current process's register state into its descriptor.
	current_proc->p_registers = *reg;
	reg = &current_proc->p_registers;

	// Show the state of memory
	show_physical_memory();

	switch (reg->reg_trapno) {

	case TRAP_SYS_PAGE_ALLOC:
		if (reg->reg_eax >= MAX_ADDRESS)
			panic("user process asks for page at address %x", reg->reg_eax);
		reg->reg_eax = page_alloc(current_proc, reg->reg_eax);
		run(current_proc);

	case TRAP_SYS_RANDOM:
		seed += read_cycle_counter();
		seed = (seed * 2147001325) + 715136305;
		reg->reg_eax = seed;
		run(current_proc);
		
	case TRAP_SYS_YIELD:
		schedule();

	case TRAP_PAGEFAULT: {
		uint32_t page_fault_address = rcr2();
		console_printf(24, 0, "%CProcess %d page fault at %x!", 0x0C00,
			       current_proc->p_pid, page_fault_address);
		current_proc->p_runnable = 0;
		schedule();
	}
		
	default:
		panic("BAD TRAP %d FROM %x", reg->reg_trapno, reg->reg_eip);
		while (1)
			/* do nothing */;

	}
}


/*****************************************************************************
 * run
 *
 *   Run the application with the supplied process descriptor.
 *   This means to reload all relevant registers from the descriptor's
 *   p_registers member, using the 'popal', 'popl', and 'iret'
 *   instructions.
 *
 *****************************************************************************/

void
run(process_t *process)
{
	current_proc = process;

	asm volatile("movl %0,%%esp\n\t"
		     "popal\n\t"
		     "popl %%es\n\t"
		     "popl %%ds\n\t"
		     "addl $8, %%esp\n\t"
		     "iret"
		     : : "g" (&process->p_registers) : "memory");

	while (1)
		/* do nothing */;
}



/*****************************************************************************
 * schedule
 *
 *   This is the weensy process scheduler; it runs the next runnable process,
 *   or if it doesn't find a runnable process, spins forever.
 *
 *****************************************************************************/

void
schedule(void)
{
	while (1) {
		int next_pid = (current_proc->p_pid + 1) % NPROCESSES;
		current_proc = &processes[next_pid];
		if (current_proc->p_runnable)
			run(current_proc);
	}
}
