#ifndef WEENSYOS_X86SYNC_H
#define WEENSYOS_X86SYNC_H
#include "types.h"

/*****************************************************************************
 * test_and_set(addr, val)
 *
 *   Executes the following as a single atomic operation:
 *
 *	uintptr_t oldval = *addr;
 *	*addr = val;
 *	return oldval;
 *
 *****************************************************************************/

static inline uintptr_t test_and_set(void *addr, uintptr_t val) __attribute__((always_inline));

static inline uintptr_t
test_and_set(void *addr, uintptr_t val)
{
	asm volatile("xchgl %0, %1"
		     : "=a" (val)
		     : "m" (*(uint32_t *)addr), "0" (val)
		     : "memory");
	return val;
}


/*****************************************************************************
 * compare_and_swap(addr, oldval, newval)
 *
 *   Executes the following as a single atomic operation:
 *
 *	if (*addr == oldval) {
 *	    *addr = newval;
 *	    return 1;
 *	} else
 *	    return 0;
 *
 *****************************************************************************/

static inline int compare_and_swap(void *addr, uintptr_t oldval, uintptr_t newval) __attribute__((always_inline));

static inline int
compare_and_swap(void *addr, uintptr_t oldval, uintptr_t newval)
{
	asm volatile("lock cmpxchg %1, %2\n\t"
		     "setz %b0"
		     : "=a" (oldval)
		     : "q" (newval), "m" (*(uint32_t *)addr), "0" (oldval)
		     : "memory");
	return (uint8_t) oldval;
}



/*****************************************************************************
 * fetch_and_add(addr, delta)
 *
 *   Executes the following as a single atomic operation:
 *
 *	uint32_t oldval = *addr;
 *	*addr += delta;
 *	return oldval;
 *
 *****************************************************************************/

static inline uint32_t fetch_and_add(void *addr, uint32_t delta) __attribute__((always_inline));

static inline uint32_t
fetch_and_add(void *addr, uint32_t delta)
{
	asm volatile("lock xaddl %0, %1"
		     : "=a" (delta)
		     : "m" (*(uint32_t *)addr), "0" (delta)
		     : "memory");
	return delta;
}

#endif /* !WEENSYOS_X86SYNC_H */
